#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import os
import pathlib

from typing import TYPE_CHECKING

import numpy
import polars
from peewee import JOIN, fn

from sdsstools.time import get_sjd

from too import log
from too.tools import match_fields
from too.validate import add_bright_limits_columns


if TYPE_CHECKING:
    from sdssdb.connection import PeeweeDatabaseConnection


__all__ = ["dump_targets_to_parquet", "dump_sdss_id_tables"]


def dump_targets_to_parquet(
    observatory: str,
    path: os.PathLike | str,
    database: PeeweeDatabaseConnection | None = None,
):
    """Dumps the ToO targets to a Parquet file.

    Parameters
    ----------
    observatory
        The observatory for which to select targets.
    path
        The path where to save the Parquet file.
    database
        The database connection. Used to check that the models are connected to
        the same database.

    """

    observatory = observatory.upper()
    if observatory not in ["APO", "LCO"]:
        raise ValueError("Invalid observatory.")

    os.environ["OBSERVATORY"] = observatory.upper()

    from sdssdb.peewee.sdss5db import catalogdb, opsdb

    path = pathlib.Path(path)

    # Check that the model connection is the same. This should be the case as
    # long as database has been created using connect_to_database(), which will
    # update the models.
    if database and database != catalogdb.Catalog._meta.database:
        raise ValueError("Mismatched database connection between too and models.")

    if database is None:
        raise ValueError("A database connection must be provided.")

    catalogdb.Catalog.bind(database)
    catalogdb.CatalogToToO_Target.bind(database)
    catalogdb.ToO_Target.bind(database)
    catalogdb.ToO_Metadata.bind(database)
    Catalog = catalogdb.Catalog
    c2too = catalogdb.CatalogToToO_Target
    TooTarget = catalogdb.ToO_Target
    TooMeta = catalogdb.ToO_Metadata

    opsdb.AssignmentToFocal.bind(database)
    Assn2Focal = opsdb.AssignmentToFocal
    Configuration = opsdb.Configuration

    columns = (
        TooTarget.too_id,
        TooTarget.fiber_type,
        Catalog.catalogid,
        TooTarget.sdss_id,
        TooTarget.ra,
        TooTarget.dec,
        TooTarget.pmra,
        TooTarget.pmdec,
        TooTarget.epoch,
        TooTarget.program,
        TooMeta.g_mag,
        TooMeta.r_mag,
        TooMeta.i_mag,
        TooMeta.z_mag,
        TooMeta.h_mag,
        TooMeta.gaia_g_mag,
        TooMeta.gaia_bp_mag,
        TooMeta.gaia_rp_mag,
        TooMeta.optical_prov,
        TooMeta.delta_ra,
        TooMeta.delta_dec,
        TooMeta.can_offset,
        TooMeta.inertial,
        TooMeta.sky_brightness_mode,
        TooMeta.n_exposures,
        TooMeta.active,
        TooMeta.observe_from_mjd,
        TooMeta.observe_until_mjd,
    )

    mjd_now = get_sjd(observatory)

    targets = (
        TooTarget.select(
            *columns,
            fn.array_agg(Configuration.epoch).alias("jd"),
        )
        .join(TooMeta, on=(TooMeta.too_id == TooTarget.too_id))
        .switch(TooTarget)
        .join(c2too)
        .join(Catalog)
        .join(
            Assn2Focal,
            JOIN.LEFT_OUTER,
            on=(Assn2Focal.catalogid == Catalog.catalogid),
        )
        .join(Configuration, JOIN.LEFT_OUTER)
        .where(
            TooMeta.active,
            (
                TooMeta.observe_until_mjd.is_null()
                | (TooMeta.observe_until_mjd >= mjd_now)
            ),
        )
        .group_by(*columns)
        .dicts()
    )

    for t in targets:
        t["jd"] = [] if t["jd"] == [None] else t["jd"]

    dataframe = polars.from_dicts(
        list(targets),
        schema={
            "too_id": polars.Int64,
            "fiber_type": polars.String,
            "catalogid": polars.Int64,
            "sdss_id": polars.Int64,
            "program": polars.String,
            "ra": polars.Float64,
            "dec": polars.Float64,
            "pmra": polars.Float32,
            "pmdec": polars.Float32,
            "epoch": polars.Float32,
            "g_mag": polars.Float32,
            "r_mag": polars.Float32,
            "i_mag": polars.Float32,
            "z_mag": polars.Float32,
            "h_mag": polars.Float32,
            "gaia_g_mag": polars.Float32,
            "gaia_bp_mag": polars.Float32,
            "gaia_rp_mag": polars.Float32,
            "optical_prov": polars.String,
            "delta_ra": polars.Float32,
            "delta_dec": polars.Float32,
            "can_offset": polars.Boolean,
            "inertial": polars.Boolean,
            "sky_brightness_mode": polars.String,
            "n_exposures": polars.Int32,
            "observe_from_mjd": polars.Int32,
            "observe_until_mjd": polars.Int32,
            "jd": polars.List(polars.Float32),
        },
    )

    df_field = match_fields(dataframe, database)
    df_observatory = df_field.filter(polars.col.observatory == observatory)

    hist = df_observatory["jd"].to_numpy()
    start = df_observatory["observe_from_mjd"].to_numpy()
    n_done = [len(numpy.where(h > s)[0]) for h, s in zip(hist, start)]
    n_requested = df_observatory["n_exposures"].to_numpy()
    elligible = df_observatory.filter(numpy.greater(n_requested, n_done)).sort("too_id")

    log.info(
        f"{observatory}: found {len(elligible)} elligible "
        f"targets of {len(targets)} total."
    )

    if len(elligible) > 0:
        # Add bright neighbour and magnitude limit check columns.
        log.info(f"{observatory}: adding bright neighbour and magnitude limit columns.")
        bn = add_bright_limits_columns(elligible, database, observatories=[observatory])
    else:
        log.warning(f"{observatory}: no elligible targets found.")
        bn = elligible.clone()

    log.debug(f"Saving data to {path!s}.")
    bn.drop_in_place("jd")
    bn.write_parquet(path)

    return bn


def dump_sdss_id_tables(
    last_updated: str,
    database: PeeweeDatabaseConnection,
    root: pathlib.Path | str = ".",
    overwrite: bool = False,
):
    """Dumps the SDSS ID tables to CSV files.

    Parameters
    ----------
    last_updated
        The value of the ``last_updated`` column in ``sandbox.sdss_id_stacked``
        to filter the data.
    database
        The database connection.
    root
        The root directory where to save the CSV files.
    overwrite
        Whether to overwrite the CSV files if they already exist.

    """

    root = pathlib.Path(root)
    root.mkdir(exist_ok=True, parents=True)

    sdss_id_flat_path = root / "sdss_id_flat.csv"
    sdss_id_stacked_path = root / "sdss_id_stacked.csv"

    if (sdss_id_flat_path.exists() or sdss_id_stacked_path.exists()) and not overwrite:
        raise FileExistsError("SDSS ID CSV files already exist.")

    sdss_id_stacked_data = polars.read_database(
        f"SELECT * FROM sandbox.sdss_id_stacked WHERE last_updated = {last_updated!r}",
        database,
        schema_overrides={
            "sdss_id": polars.Int64,
            "catalogid21": polars.Int64,
            "catalogid25": polars.Int64,
            "catalogid31": polars.Int64,
        },
    )

    sdss_id_flat_data = polars.read_database(
        f"""SELECT flat.* FROM sandbox.sdss_id_flat flat
                JOIN sandbox.sdss_id_stacked stacked ON stacked.sdss_id = flat.sdss_id
                WHERE stacked.last_updated = {last_updated!r}
        """,
        database,
    )

    sdss_id_stacked_data.write_csv(sdss_id_stacked_path)
    sdss_id_flat_data.write_csv(sdss_id_flat_path)

    return sdss_id_stacked_path, sdss_id_flat_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-16
# @Filename: carton.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib
import warnings

from typing import TYPE_CHECKING

from too import log


if TYPE_CHECKING:
    from sdssdb.connection import PeeweeDatabaseConnection


__all__ = ["run_too_carton", "update_sdss_id_tables"]


TOO_TARGET_PLAN = "1.1.0"
TOO_TAG = "too"


def run_too_carton():
    """Runs the ToO carton. Database connection must have already been set."""

    from target_selection import log as ts_log
    from target_selection.exceptions import TargetSelectionImportWarning

    ts_log.setLevel(40)  # Disable target_selection logging but allow exceptions.
    warnings.simplefilter("ignore", TargetSelectionImportWarning)

    from target_selection.cartons.too import ToO_Carton  # Slow import

    too_carton = ToO_Carton(TOO_TARGET_PLAN)
    too_carton.tag = TOO_TAG
    too_carton.log = log

    too_carton.run(overwrite=True)
    too_carton.load(mode="append")


def backup_sdss_id_tables(
    database: PeeweeDatabaseConnection,
    tables=[
        "sdss_id_flat",
        "sdss_id_flat_addendum",
        "sdss_id_stacked",
        "sdss_id_stacked_addendum",
    ],
    schema="sandbox",
    outdir: pathlib.Path | str = ".",
    suffix: str = "",
):
    """Backs up the SDSS ID tables.

    Parameters
    ----------
    database
        The database connection.
    tables
        The tables to backup. Each table is backed up as a separate file in
        ``outdir`` with the format ``<schema>_<table_name>_<suffix>.csv``.
    schema
        The schema where the tables are located.
    outdir
        The output directory for the backup files. Defaults to the current
        directory.
    suffix
        A suffix to add to the backup files.

    """

    outdir = pathlib.Path(outdir).absolute()
    outdir.mkdir(parents=True, exist_ok=True)

    if suffix:
        suffix = f"_{suffix}"

    assert database.connected, "Database connection must be established."

    for table in tables:
        cursor = database.cursor()
        with open(outdir / f"{schema}_{table}{suffix}.csv", "w") as file:
            cursor.copy_expert(f"COPY {schema}.{table} TO STDOUT WITH CSV HEADER", file)


def update_sdss_id_tables(database: PeeweeDatabaseConnection):
    """Updates the SDSS ID tables."""

    from target_selection import log as ts_log
    from target_selection.sdss_id import (
        SdssIdFlatAddendum,
        SdssIdStackedAddendum,
        TempCatalogidV21,
        TempCatalogidV25,
        TempCatalogidV31,
        append_to_sdss_id,
    )
    from target_selection.sdss_id.append_to_sdss_id import database as append_database
    from target_selection.sdss_id.create_catalogidx_to_catalogidy import (
        database as cidx_database,
    )

    # Override log
    if ts_log.sh in ts_log.handlers:
        ts_log.removeHandler(ts_log.sh)
    ts_log.addHandler(log.sh)
    ts_log.sh = log.sh

    # The sdss_id code uses whichever database profile is default for the machine.
    # This is necessary to ensure that the database that we pass to this function
    # is actually used.
    append_database.connect(dbname=database.dbname, **database.connect_params)
    append_database._config.update(**database.connect_params)

    cidx_database.connect(dbname=database.dbname, **database.connect_params)
    cidx_database._config.update(**database.connect_params)

    log.info("Updating SDSS ID tables.")

    apend_inst = append_to_sdss_id.AppendToTables(
        database,
        individual_table="catalogdb.catalog_to_too_target",
    )

    output_name = apend_inst.run_MetaXMatch(database)

    apend_inst.create_temp_catalogid_lists(database, output_name)
    apend_inst.create_sdss_id_stacked_addendum(database, output_name)
    apend_inst.add_to_sdss_id_stacked(database)
    apend_inst.create_sdss_id_flat_addendum(database)
    apend_inst.add_to_sdss_id_flat(database)

    # Drop temporary tables after successful update.
    append_database.drop_tables(
        [
            TempCatalogidV21,
            TempCatalogidV25,
            TempCatalogidV31,
            SdssIdStackedAddendum,
            SdssIdFlatAddendum,
        ]
    )

    log.info("SDSS ID tables have been updated.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Ilija Medan (ilija.medan@vanderbilt.edu)
# @Date: 2024-04-02
# @Filename: validate.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import collections
import os
import warnings

from typing import TYPE_CHECKING, Literal, Tuple

import astropy.units as u
import numpy as np
import polars
from astropy.coordinates import SkyCoord
from astropy.time import Time
from erfa import ErfaWarning
from polars import selectors

from too import log
from too.datamodel import mag_columns, too_dtypes
from too.exceptions import ValidationError
from too.tools import match_fields


if TYPE_CHECKING:
    from sdssdb import PeeweeDatabaseConnection


__all__ = ["validate_too_targets", "add_bright_limits_columns"]


# load enviornment variable with path to healpix maps
BN_HEALPIX = os.getenv("BN_HEALPIX")


def check_assign_mag_limit(
    mag_metric_min,
    mag_metric_max,
    assign_mag,
):  # pragma: no cover
    """
    Checks the if magnitude of one assignment agrees with
    design mode for some instrument and carton class

    Parameters
    ----------
    mag_metric_min: float
        The minimum magitude for the specific DesignMode.

    mag_metric_min: float
        The maximum magitude for the specific DesignMode.

    assign_mag: float
        The magntiude of the assignment being checkaged against
        the DesignMode.

    Returns
    -------
    targ_check: boolean
        True of the assignment magntiude passed the DesignMode
        for the magnitude, False if not.

    complete_check: str
        'COMPLETE' if assign_mag is value, 'INCOMPLETE' if Null.

    """

    # set default values
    complete_check = "COMPLETE"
    targ_check = False

    # do checks
    # define Null cases for targetdb.magnitude table
    cases = [-999, -9999, 999, 0.0, np.nan, 99.9, None]
    if assign_mag in cases or np.isnan(assign_mag):
        complete_check = "INCOMPLETE"
        # set True, no mag is not a fail
        targ_check = True
    # check when greater than and less than
    elif mag_metric_min != -999.0 and mag_metric_max != -999.0:
        if mag_metric_min < assign_mag < mag_metric_max:
            targ_check = True
    # check when just greater than
    elif mag_metric_min != -999.0:
        if assign_mag > mag_metric_min:
            targ_check = True
    # check when less than
    else:
        if assign_mag < mag_metric_max:
            targ_check = True
    return targ_check, complete_check


def allDesignModes(database: PeeweeDatabaseConnection):
    """Function to return a dictionary with all design modes.

    Adapted from the original function in ``mugatu.designmode``.

    Parameters
    ----------
    database : PeeweeDatabaseConnection
        The database connection to use to query the design modes.

    Comments
    --------
    Returns an OrderedDict

    If filename is not provided, it reads from that file name
    and the given extension. If not, it reads from the database.

    If neither the filename is given nor the db is available, it
    returns None.

    """

    from sdssdb.peewee.sdss5db.targetdb import DesignMode as DesignModeDB

    desmodes = DesignModeDB.select()
    dmd = collections.OrderedDict()
    for desmode in desmodes:
        dmd[desmode.label] = DesignMode(database, label=desmode.label)
    return dmd


class DesignMode:
    """Class to store parameters for a design mode.

    Adapted from the original class in ``mugatu.designmode``.

    Parameters
    ----------
    database : PeeweeDatabaseConnection
        The database connection to use to query the design modes.
    label : str
        Label for design mode

    Attributes
    ----------
    n_stds_min: dict
        Dictionary with the minimum number of standards for a design
        for each instrument ('APOGEE' and 'BOSS').

    min_stds_fovmetric: dict
        Dictionary wih the FOV metric for the standards in a design
        for each instrument ('APOGEE' and 'BOSS').
        The FOV metric is described by three parameters, the
        nth neighbor to get distances to, the percentle distance
        to calculate and the distance to compare against for validation
        (in mm).

    stds_mags: dict
        Dictionary for the min/max magnitude for the standards in a
        design for each instrument ('APOGEE' and 'BOSS'). Indexes
        correspond to magntidues: [g, r, i, z, bp, gaia_g, rp, J, H, K].

    bright_limit_targets: dict
        Dictionary for the min/max magnitude for the science targets
        in adesign for each instrument ('APOGEE' and 'BOSS'). Indexes
        correspond to magntidues: [g, r, i, z, bp, gaia_g, rp, J, H, K].

    sky_neighbors_targets: dict
        Dictionary for the parameters used to check distance between
        skies and all possible sources in field for each instrument
        ('APOGEE' and 'BOSS'). Distances to targets (r, in arcseconds)
        must be r > R_0 * (lim - mags) ** beta, where mags is G band
        for BOSS and H band for APOGEE. Indexes correspond to:
        [R_0, beta, lim]

    trace_diff_targets: dict
        Dictionary for the maximum magnitude difference allowed between
        fibers next to each ther on the chip for each instrument
        ('APOGEE' and 'BOSS'). Here the magntidue difference is checked
        in the G band for BOSS and H band for APOGEE.

    """

    def __init__(self, database: PeeweeDatabaseConnection, label: str | None = None):
        from sdssdb.peewee.sdss5db.targetdb import DesignMode as DesignModeDB

        DesignModeDB._meta.database(database)

        if label is not None:
            self.fromdb(label=label)
        return

    def fromdb(self, label: str):
        """Read in parameters for design mode from db.

        Parameters
        ----------

        label : str
            name of design mode

        """

        from sdssdb.peewee.sdss5db.targetdb import DesignMode as DesignModeDB

        self.desmode_label = label

        desmode = DesignModeDB.select().where(DesignModeDB.label == label)[0]

        self.n_skies_min = {}
        self.n_skies_min["BOSS"] = desmode.boss_skies_min
        self.n_skies_min["APOGEE"] = desmode.apogee_skies_min

        self.min_skies_fovmetric = {}
        self.min_skies_fovmetric["BOSS"] = desmode.boss_skies_fov
        self.min_skies_fovmetric["APOGEE"] = desmode.apogee_skies_fov

        self.n_stds_min = {}
        self.n_stds_min["BOSS"] = desmode.boss_stds_min
        self.n_stds_min["APOGEE"] = desmode.apogee_stds_min

        self.min_stds_fovmetric = {}
        self.min_stds_fovmetric["BOSS"] = desmode.boss_stds_fov
        self.min_stds_fovmetric["APOGEE"] = desmode.apogee_stds_fov

        self.stds_mags = {}
        self.stds_mags["BOSS"] = np.zeros(
            (len(desmode.boss_stds_mags_min), 2),
            dtype=np.float64,
        )
        self.stds_mags["BOSS"][:, 0] = desmode.boss_stds_mags_min
        self.stds_mags["BOSS"][:, 1] = desmode.boss_stds_mags_max
        self.stds_mags["APOGEE"] = np.zeros(
            (len(desmode.apogee_stds_mags_min), 2),
            dtype=np.float64,
        )
        self.stds_mags["APOGEE"][:, 0] = desmode.apogee_stds_mags_min
        self.stds_mags["APOGEE"][:, 1] = desmode.apogee_stds_mags_max

        self.bright_limit_targets = {}
        self.bright_limit_targets["BOSS"] = np.zeros(
            (len(desmode.boss_bright_limit_targets_min), 2),
            dtype=np.float64,
        )
        self.bright_limit_targets["BOSS"][:, 0] = desmode.boss_bright_limit_targets_min
        self.bright_limit_targets["BOSS"][:, 1] = desmode.boss_bright_limit_targets_max

        dmode_apogee_blimit_min = desmode.apogee_bright_limit_targets_min
        dmode_apogee_blimit_max = desmode.apogee_bright_limit_targets_max

        self.bright_limit_targets["APOGEE"] = np.zeros(
            (len(dmode_apogee_blimit_min), 2),
            dtype=np.float64,
        )
        self.bright_limit_targets["APOGEE"][:, 0] = dmode_apogee_blimit_min
        self.bright_limit_targets["APOGEE"][:, 1] = dmode_apogee_blimit_max

        self.sky_neighbors_targets = {}
        self.sky_neighbors_targets["BOSS"] = desmode.boss_sky_neighbors_targets
        self.sky_neighbors_targets["APOGEE"] = desmode.apogee_sky_neighbors_targets

        self.trace_diff_targets = {}
        self.trace_diff_targets["APOGEE"] = desmode.apogee_trace_diff_targets

        return


def magnitude_array(targets: polars.DataFrame) -> np.ndarray:  # pragma: no cover
    """
    create the magnitude array for the targets
    Parameters
    ----------
    targets: polars.DataFrame
        DataFrame with the ToO target information
    Returns
    -------
    magnitudes: np.array
        Magnitude array for the targets. Array is of size (N, 10),
        where columns correspond to
        g, r, i, z, bp, gaia_g, rp, J, H, K band
        magnitudes.

    """

    magnitudes = targets[
        [
            "g_mag",
            "r_mag",
            "i_mag",
            "z_mag",
            "gaia_bp_mag",
            "gaia_g_mag",
            "gaia_rp_mag",
        ]
    ].to_numpy()

    # add in the missing 2MASS mags as nan
    magnitudes = np.column_stack(
        (
            magnitudes,
            np.zeros(len(targets)) + np.nan,
            targets["h_mag"],
            np.zeros(len(targets)) + np.nan,
        )
    )
    return magnitudes


def calculate_offsets(
    targets: polars.DataFrame,
    design_mode: str,
    modes: dict,
    offset_min_skybrightness: float = 0.0,
    observatory: str = "APO",
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:  # pragma: no cover
    """
    Calculate the offsets for the targets

    Parameters
    ----------
    targets: polars.DataFrame
        DataFrame with the ToO target information

    design_mode: str
        The design_mode to run the validation for

    offset_min_skybrightness: float
        Minimum skybrightness targets will be offset

    observatory: str
        Observatory where observation is taking place, either
        'LCO' or 'APO'.

    Return
    ------
    delta_ra: np.array
        offset in the R.A. direction

    delta_dec: np.array
        offset in the Decl. direction

    offset_flag: np.array
        flags associated with the offseting

    """

    from coordio.utils import Moffat2dInterp, object_offset

    # add function for offseting
    fmagloss = Moffat2dInterp()

    delta_ra = np.zeros(len(targets), dtype=float)
    delta_dec = np.zeros(len(targets), dtype=float)
    offset_flag = np.zeros(len(targets), dtype=int)
    if "bright" in design_mode:
        boss_mag_lim = modes[design_mode].bright_limit_targets["BOSS"][:, 0]
        lunation = "bright"
        skybrightness = 1.0
    else:
        boss_mag_lim = modes[design_mode].bright_limit_targets["BOSS"][:, 0]
        lunation = "dark"
        skybrightness = 0.35
    apogee_mag_lim = modes[design_mode].bright_limit_targets["APOGEE"][:, 0]

    magnitudes = magnitude_array(targets)

    ev_boss = targets["fiber_type"] == "BOSS"
    res = object_offset(
        magnitudes[ev_boss, :],
        boss_mag_lim,
        lunation,
        "Boss",
        observatory,
        fmagloss=fmagloss,
        can_offset=targets["can_offset"].to_numpy()[ev_boss],
        skybrightness=skybrightness,
        offset_min_skybrightness=offset_min_skybrightness,
    )
    delta_ra[ev_boss] = res[0]
    delta_dec[ev_boss] = res[1]
    offset_flag[ev_boss] = res[2]

    ev_apogee = targets["fiber_type"] == "APOGEE"
    res = object_offset(
        magnitudes[ev_apogee, :],
        apogee_mag_lim,
        lunation,
        "Apogee",
        observatory,
        fmagloss=fmagloss,
        can_offset=targets["can_offset"].to_numpy()[ev_apogee],
        skybrightness=skybrightness,
        offset_min_skybrightness=offset_min_skybrightness,
    )
    delta_ra[ev_apogee] = res[0]
    delta_dec[ev_apogee] = res[1]
    offset_flag[ev_apogee] = res[2]
    return delta_ra, delta_dec, offset_flag


def bn_validation(
    targets: polars.DataFrame,
    design_mode: str,
    modes: dict,
    observatory: str = "APO",
) -> np.ndarray:  # pragma: no cover
    """
    Validate a ToO to see if it is too close
    to a bright neighbor. This functionm relies on the
    healpix map of bright neighbors for each designmode
    created with this script:
    https://github.com/sdss/mugatu/blob/master/bin/bright_neigh_healpix_map.py.
    To use this function, the environment variable BN_HEALPIX must be
    set to the path to these healpix maps.

    Parameters
    ----------
    targets: polars.DataFrame
        DataFrame with the ToO target information
    design_mode: str
        The design_mode to run the validation for
    modes : dict
        A dictionary of design mode to ``DesignMode`` object.
    observatory: str
        Observatory where observation is taking place, either
        'LCO' or 'APO'.

    Return
    ------
    valid_too_bn: np.array
        bright neighbor validation for specified design_mode.
        True means passes check.

    """

    from astropy_healpix import HEALPix

    from coordio.utils import _offset_radec

    log.debug(
        f"Running bright neighbour validation for observatory {observatory} "
        f"and design mode {design_mode}"
    )

    BN_HEALPIX = os.getenv("BN_HEALPIX")
    if BN_HEALPIX is None:
        raise ValueError("Environment variable BN_HEALPIX not set.")

    if not os.path.exists(BN_HEALPIX):
        raise ValueError(f"Path $BN_HEALPIX={BN_HEALPIX} does not exist.")

    # load the healpix indicies for the designmode
    bn_maps_boss_file = f"{BN_HEALPIX}/{design_mode}_boss_bn_healpix.parquet"
    if not os.path.exists(bn_maps_boss_file):
        raise FileNotFoundError(f"File {bn_maps_boss_file} does not exist.")
    bn_maps_boss = polars.scan_parquet(bn_maps_boss_file)

    bn_maps_apogee_file = f"{BN_HEALPIX}/{design_mode}_apogee_bn_healpix.parquet"
    if not os.path.exists(bn_maps_apogee_file):
        raise FileNotFoundError(f"File {bn_maps_apogee_file} does not exist.")
    bn_maps_apogee = polars.scan_parquet(bn_maps_apogee_file)

    # create the correct nside healpix object
    hp = HEALPix(nside=2**18, order="ring", frame="icrs")

    # calculate the offsets
    delta_ra, delta_dec, _ = calculate_offsets(
        targets,
        design_mode,
        modes=modes,
        observatory=observatory,
    )
    # get the new ra and dec
    # still need proper motion correction?
    ra_off, dec_off = _offset_radec(
        ra=targets["ra"].to_numpy(),
        dec=targets["dec"].to_numpy(),
        delta_ra=delta_ra,
        delta_dec=delta_dec,
    )
    # get proper motions
    pmra = targets["pmra"].to_numpy()
    pmdec = targets["pmdec"].to_numpy()
    epoch = targets["epoch"].to_numpy()
    # deal with nulls
    ev_pm_null = np.isnan(pmra) | np.isnan(pmdec)
    pmra[ev_pm_null] = 0.0
    pmdec[ev_pm_null] = 0.0
    epoch[np.isnan(epoch)] = 2016.0
    # get the indicies for all targets
    coord = SkyCoord(
        ra=ra_off * u.deg,
        dec=dec_off * u.deg,
        pm_ra_cosdec=pmra * u.mas / u.yr,
        pm_dec=pmdec * u.mas / u.yr,
        obstime=Time(epoch, format="decimalyear"),
    ).apply_space_motion(Time.now())
    hp_inds = hp.skycoord_to_healpix(coord)

    assert isinstance(hp_inds, np.ndarray)

    # check if ToOs in bright neighbor healpixels
    valid_too_bn = np.zeros(len(targets), dtype=bool) + True

    ev_boss = targets["fiber_type"] == "BOSS"
    ev_apogee = targets["fiber_type"] == "APOGEE"

    hp_inds_boss = hp_inds[ev_boss]
    hp_inds_apogee = hp_inds[ev_apogee]

    # Get bright healpixels that are in our target list.
    invalid_hp_boss = bn_maps_boss.filter(polars.col.data.is_in(hp_inds_boss))
    invalid_hp_apogee = bn_maps_apogee.filter(polars.col.data.is_in(hp_inds_apogee))

    # opposite as True == valid target
    valid_too_bn[ev_boss] = ~np.isin(hp_inds[ev_boss], invalid_hp_boss.collect())
    valid_too_bn[ev_apogee] = ~np.isin(hp_inds[ev_apogee], invalid_hp_apogee.collect())

    return valid_too_bn


def mag_limit_check(
    magnitudes: np.ndarray,
    can_offset: np.ndarray,
    offset_flag: np.ndarray,
    mag_metric: np.ndarray,
) -> np.ndarray:
    """
    perform a specific magnitude check
    Parameters
    ----------
    magnitudes: np.array
        Magnitude array for the targets. Array is of size (N, 10),
        where columns correspond to
        g, r, i, z, bp, gaia_g, rp, J, H, K band
        magnitudes.
    can_offset: np.array
        If target can be offset or not
    offset_flag: np.array
        offset_flags from offset function
    mag_metric: np.array
        the magnitude limits for the specific designmode
        and intrument
    Return
    ------
    valid_too_mag_lim: np.array
        magnitude limit validation for specified design_mode.
        True means passes check.
    """
    valid_too_mag_lim = np.zeros(len(magnitudes), dtype=bool)
    check_inds = []
    for i in range(mag_metric.shape[0]):
        if mag_metric[i][0] != -999.0 or mag_metric[i][1] != -999.0:
            check_inds.append(i)

    # run checks
    for i in range(len(valid_too_mag_lim)):
        # don't do check and make true if offset target
        if can_offset[i] and offset_flag[i] == 0:
            valid_too_mag_lim[i] = True
        else:
            # check in each band that has check defined
            targ_check = np.zeros(len(check_inds), dtype=bool)
            for j, ind in enumerate(check_inds):
                # check the magntiude for this assignment
                targ_check[j], _ = check_assign_mag_limit(
                    mag_metric[ind][0],
                    mag_metric[ind][1],
                    magnitudes[i][ind],
                )
            # if all True, then passes
            if np.all(targ_check):
                valid_too_mag_lim[i] = True
    return valid_too_mag_lim


def mag_lim_validation(
    targets: polars.DataFrame,
    design_mode: str,
    modes: dict,
    observatory: str = "APO",
) -> np.ndarray:
    """
    Perform a magnitude limit check on the ToOs
    Parameters
    ----------
    targets: polars.DataFrame
        DataFrame with the ToO target information
    design_mode: str
        The design_mode to run the validation for
    modes : dict
        A dictionary of design mode to ``DesignMode`` object.
    observatory: str
        Observatory where observation is taking place, either
        'LCO' or 'APO'.
    Return
    ------
    valid_too_mag_lim: np.array
        magnitude limit validation for specified design_mode.
        True means passes check.
    """

    log.debug(
        f"Running magnitude limit check for observatory {observatory} "
        f"and design mode {design_mode}"
    )

    valid_too_mag_lim = np.zeros(len(targets), dtype=bool)

    # get the magnitude array
    magnitudes = magnitude_array(targets)

    # get offset flags
    # calculate the offsets
    _, _, offset_flag = calculate_offsets(
        targets,
        design_mode,
        modes,
        observatory=observatory,
    )

    # do check for BOSS
    ev_boss = targets["fiber_type"] == "BOSS"
    valid_too_mag_lim[ev_boss] = mag_limit_check(
        magnitudes[ev_boss, :],
        targets["can_offset"].to_numpy()[ev_boss],
        offset_flag[ev_boss],
        modes[design_mode].bright_limit_targets["BOSS"],
    )

    # do check for APOGEE
    ev_apogee = targets["fiber_type"] == "APOGEE"
    valid_too_mag_lim[ev_apogee] = mag_limit_check(
        magnitudes[ev_apogee, :],
        targets["can_offset"].to_numpy()[ev_apogee],
        offset_flag[ev_apogee],
        modes[design_mode].bright_limit_targets["APOGEE"],
    )

    return valid_too_mag_lim


def validate_too_targets(
    targets: polars.DataFrame,
    database: PeeweeDatabaseConnection | None = None,
    bright_limit_checks: bool = False,
    bright_limit_mode: Literal["error", "discard"] = "discard",
):
    """Validates a list of ToO targets.

    Checks the following conditions:

    - The dataframe schema matches the ``too_target`` schema.
    - The ``too_id`` column is unique.
    - ``ra`` and ``dec`` are present and valid.
    - At least one of the magnitude columns is present.
    - The number of exposures is set and valid.
    - The ``active`` column is set.

    If ``bright_limit_checks`` is ``True``, the function will also run the bright
    neighbour and bright limit checks and will fail if any of the targets do not pass
    the checks. When ``bright_limit_mode`` is set to 'error', the function will raise
    a ``ValidationError`` if any of the targets do not pass the checks. Otherwise,
    invalid rows will be discarded.

    """

    from sdsstools import get_sjd

    n_targets = targets.height

    if targets.schema != too_dtypes:
        raise ValidationError("Invalid schema for ToO targets.")

    if targets.unique("too_id").height != n_targets:
        raise ValidationError("Duplicate too_id in ToO targets.")

    if targets["too_id"].has_nulls():
        raise ValidationError("Null too_id found in ToO targets.")

    if targets["sky_brightness_mode"].has_nulls():
        raise ValidationError("Null sky_brightness_mode found in ToO targets.")

    targets_coords = targets.select(["ra", "dec"]).drop_nulls()
    if len(targets_coords) < n_targets:
        raise ValidationError("Null ra/dec found in ToO targets.")

    bad_ra = (targets_coords["ra"] >= 360) | (targets_coords["ra"] < 0)
    bad_dec = (targets_coords["dec"] >= 90) | (targets_coords["dec"] <= -90)
    if bad_ra.any() or bad_dec.any():
        raise ValidationError("Invalid ra or dec found in ToO targets.")

    if targets["n_exposures"].is_null().any():
        raise ValidationError("Null 'n_exposures' column values found in ToO targets.")

    if targets["active"].is_null().any():
        raise ValidationError("Null 'active' column values found in ToO targets.")

    # Get magnitude columns and filtered rows that do not have any value set.
    # This is equivalent to Pandas .drop(axis=1). In Polars is a bit harder. See
    # https://github.com/pola-rs/polars/issues/1613
    mag_data = targets[mag_columns]
    any_mags = mag_data.with_columns(
        no_mags=polars.fold(
            True,
            lambda acc, s: acc & s.is_null(),
            polars.all(),
        )
    )
    if len(any_mags.filter(polars.col.no_mags.not_())) < n_targets:
        raise ValidationError(
            "ToOs found with missing magnitudes. "
            "At least one magnitude value is required."
        )

    if len(set(targets["fiber_type"].unique()) - set(["APOGEE", "BOSS"])) > 0:
        raise ValidationError(
            "Invalid fiber_type values. Valid values are 'APOGEE' and 'BOSS'."
        )

    # Check can_offset is a boolean and is set.
    if targets["can_offset"].is_null().any():
        raise ValidationError("Null 'can_offset' column values found in ToO targets.")

    # Check that the sky_brightness_mode value are valid.
    valid_brightness_modes = ["bright", "dark", "either"]

    targets_invalid_brightness_mode = targets.filter(
        polars.col.sky_brightness_mode.is_in(valid_brightness_modes).not_()
    )
    if len(targets_invalid_brightness_mode) > 0:
        raise ValidationError("Invalid sky_brightness_mode values found.")

    invalid_bright_limits = set()
    if bright_limit_checks:
        if database is None:
            raise ValueError("Database connection is required for bright limit checks.")

        # Validate magnitude limits and bright neighbours.
        log.info("Running bright neighbour and magnitude limit checks.")
        bn_targets = add_bright_limits_columns(targets, database)

        # For the sky brightness mode of the target,
        # check if the bright neighbour and magnitude limit.
        for bmode in valid_brightness_modes:
            bm_targets = bn_targets.filter(polars.col.sky_brightness_mode == bmode)
            bn_cols = bm_targets.select(selectors.starts_with(f"bn_{bmode}"))

            if len(bm_targets) == 0 or len(bn_cols) == 0:
                continue

            bn_invalid = bm_targets.filter(~bn_cols.fold(lambda x, y: x & y))

            mag_lim_cols = bm_targets.select(selectors.starts_with(f"mag_lim_{bmode}"))
            mag_lim_invalid = bm_targets.filter(~mag_lim_cols.fold(lambda x, y: x & y))

            if len(bn_invalid) > 0 or len(mag_lim_invalid) > 0:
                if bright_limit_mode == "error":
                    raise ValidationError(
                        f"{len(bn_invalid)} targets failed bright neighbour or "
                        "magnitude limit checks."
                    )
                else:
                    invalid_bright_limits |= set(bn_invalid["too_id"].to_list())
                    invalid_bright_limits |= set(mag_lim_invalid["too_id"].to_list())
                    log.warning(
                        f"{len(invalid_bright_limits)} targets failed bright "
                        f"neighbour or magnitude limit checks and have been discarded."
                    )

    # Fill some optional columns.
    observe_from_now = polars.lit(get_sjd("APO"), dtype=polars.Int32)
    targets = targets.with_columns(
        active=polars.col.active.fill_null(True),
        inertial=polars.col.inertial.fill_null(False),
        observed=polars.col.observed.fill_null(False),
        optical_prov=polars.col.optical_prov.fill_null(""),
        delta_ra=polars.col.delta_ra.fill_null(polars.lit(0, dtype=polars.Float32)),
        delta_dec=polars.col.delta_dec.fill_null(polars.lit(0, dtype=polars.Float32)),
        n_exposures=polars.col.n_exposures.fill_null(polars.lit(1, dtype=polars.Int16)),
        priority=polars.col.priority.fill_null(polars.lit(5, dtype=polars.Int16)),
        observe_from_mjd=polars.col.observe_from_mjd.fill_null(observe_from_now),
    ).filter(polars.col.too_id.is_in(invalid_bright_limits).not_())

    return targets


def add_bright_limits_columns(
    targets: polars.DataFrame,
    database: PeeweeDatabaseConnection,
    observatories: list[str] = ["APO", "LCO"],
):
    """Runs the Mugatu-like validation for bright targets.

    Parameters
    ----------
    targets
        The ToO targets to validate.
    database
        The database connection to use to query the design modes.
    observatories
        The list of observatories to check.

    Returns
    -------
    data_frame
        A data frame with the same rows as the input ``targets`` with additional
        columns that indicate if a target passes the bright neighbour and magnitude
        limit checks. The order of the returned data frame is sorted by ``too_id``
        and may be different as that of the input one.

    """

    targets = targets.clone()

    # Get the field/observatory for each tile.
    if "observatory" not in targets.columns:
        targets = match_fields(targets, database)

    # Check that the sky_brightness_mode value are valid.
    valid_brightness_modes = ["bright", "dark"]

    # Get a list of design modes.
    design_modes: dict[str, DesignMode] = allDesignModes(database)

    # Add all columns for all design modes and set to false for now.
    targets = targets.with_columns(
        **{f"bn_{dm}_valid": polars.lit(False) for dm in design_modes},
        **{f"mag_lim_{dm}_valid": polars.lit(False) for dm in design_modes},
    )

    processed: list[polars.DataFrame] = []

    # First we split the targets by observatory and sky_brightness_mode (bright or
    # dark). Then we run the check for all the design modes that match that brightness
    # mode. We do not reject targets here but we mark if they pass the bright neighbour
    # and magnitude limit checks.
    for observatory in observatories:
        for bmode in valid_brightness_modes:
            targets_bmode = targets.filter(
                (
                    (polars.col.sky_brightness_mode == bmode)
                    | (polars.col.sky_brightness_mode == "either")
                )
                & (polars.col.observatory == observatory)
            )

            if len(targets_bmode) == 0:
                continue

            design_modes_bmode = [dm for dm in design_modes if dm.startswith(bmode)]

            for dmb in design_modes_bmode:
                try:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=ErfaWarning)

                        bn_mask = bn_validation(
                            targets_bmode,
                            dmb,
                            design_modes,
                            observatory=observatory,
                        )

                except FileNotFoundError:
                    # If there is not a BN file for this design mode, we just skip it
                    # and assume all targets pass the check.
                    bn_mask = np.ones(len(targets_bmode), dtype=bool)

                except Exception:
                    raise

                mag_lim_mask = mag_lim_validation(
                    targets_bmode,
                    dmb,
                    design_modes,
                    observatory=observatory,
                )

                targets_bmode = targets_bmode.with_columns(
                    **{
                        f"bn_{dmb}_valid": polars.Series(bn_mask),
                        f"mag_lim_{dmb}_valid": polars.Series(mag_lim_mask),
                    }
                )

            processed.append(targets_bmode)

    # Concatenate all the processed data frames. For targets that do not have a
    # corresponding design mode, the columns will be nulls. We fill those with False
    # as in principle we should not be able to observe the targets in those modes.
    processed_df = polars.concat(processed, how="diagonal")
    processed_df = processed_df.with_columns(
        selectors.starts_with("bn_").fill_null(False),
        selectors.starts_with("mag_lim_").fill_null(False),
    )

    # For targets with sky_brightness_mode='either', we will have two rows, one
    # evaluated for each brightness mode. We collapse the bn_ and mag_lim_ columns
    # and remove the duplicates.
    processed_either = (
        processed_df.filter(polars.col.sky_brightness_mode == "either")
        .with_columns(
            polars.selectors.starts_with("mag_").any(),
            polars.selectors.starts_with("bn_").any(),
        )
        .unique("too_id")
    )

    final_df = polars.concat(
        [
            processed_df.filter(polars.col.sky_brightness_mode != "either"),
            processed_either,
        ]
    )

    return final_df.sort("too_id")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-12
# @Filename: xmatch.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING, cast

import peewee
import polars

from too import log
from too.database import get_database_uri


if TYPE_CHECKING:
    from sdssdb.connection import PeeweeDatabaseConnection


__all__ = ["xmatch_too_targets"]


TOO_RUN_ID = 10
TOO_XMATCH_PLAN = "1.2.0"
TOO_XMATCH_CONFIG = {
    "1.2.0": {
        "run_id": TOO_RUN_ID,
        "query_radius": 1.0,
        "show_sql": True,
        "schema": "catalogdb",
        "output_table": "catalog",
        "start_node": "gaia_dr3_source",
        "debug": True,
        "log_path": "xmatch_{plan}.log",
        "path_mode": "config_list",
        "extra_nodes": [
            "gaia_dr3_source",
            "catalog_to_gaia_dr3_source",
            "twomass_psc",
            "catalog_to_twomass_psc",
        ],
        "version_id": 31,
        "order": ["too_target"],
        "tables": {
            "too_target": {
                "ra_column": "ra",
                "dec_column": "dec",
                "pmra_column": "pmra",
                "pmdec_column": "pmdec",
                "is_pmra_cos": "true",
                "parallax_column": "parallax",
                "has_missing_coordinates": False,
                "epoch": 2016.0,
                "allow_multiple_bests": True,
            }
        },
        "join_paths": [
            [
                "too_target",
                "gaia_dr3_source",
                "catalog_to_gaia_dr3_source",
                "catalog",
            ],
            [
                "too_target",
                "twomass_psc",
                "catalog_to_twomass_psc",
                "catalog",
            ],
        ],
        "database_options": {
            "work_mem": "10GB",
            "temp_buffers": "5GB",
            "maintenance_work_mem": "5GB",
            "enable_hashjoin": False,
        },
    }
}


def xmatch_too_targets(
    database: PeeweeDatabaseConnection,
    plan: str | None = None,
    dry_run: bool = False,
    overwrite=False,
    keep_temp: bool = False,
):
    """Performs a cross-match of the ToO targets with the SDSS catalogues.

    The routine performs the folling steps:

    - Parses the ``too_target`` table.
    - Identifies ToO entries with either a ``catalogid`` or an ``sdss_id`` and
      manually them to the ``catalog_to_too_target`` table (if they are not
      already there).
    - For the remaining ToO entries, performs a cross-match with the SDSS
      catalogues and adds the matches to the ``catalog_to_too_target`` table.
      This is accomplished calling ``target_selection.XMatch`` in "addendum"
      mode.

    Parameters
    ----------
    database
        The database connection to use.
    plan
        The cross-match plan to use. Defaults to ``TOO_XMATCH_PLAN``.
    dry_run
        Whether to load the ``catalog`` and ``catalog_to_too_target`` tables after
        the cross-match.
    overwrite
        Deletes the temporary tables if they exist.
    keep_temp
        Whether to keep the temporary tables after the cross-match.

    Returns
    -------
    model
        The catalog model where the cross-match results are stored. If
        ``load_catalog`` is ``False``, returns the model for the temporary
        table.

    """

    from sdssdb.peewee.sdss5db.catalogdb import ToO_Target
    from target_selection.xmatch import XMatchPlanner

    assert database.connected, "Database is not connected"

    too_target_schema: str = ToO_Target._meta.schema
    too_target_table_name: str = ToO_Target._meta.table_name
    catalog_to_target_table_name: str = f"catalog_to_{too_target_table_name}"

    too_rel_fqtn = f"{too_target_schema}.{catalog_to_target_table_name}"
    too_fqtn = f"{too_target_schema}.{too_target_table_name}"

    database_uri = get_database_uri(database.dbname, **database.connect_params)

    # Some basic checks.
    assert database.table_exists(
        too_target_table_name,
        schema=too_target_schema,
    ), f"Table {too_fqtn} does not exist."

    assert database.table_exists(
        catalog_to_target_table_name,
        schema=too_target_schema,
    ), f"Table {too_rel_fqtn} does not exist."

    plan_config = TOO_XMATCH_CONFIG.copy()
    plan_config["1.2.0"]["debug"] = log.sh.level

    plan = plan or TOO_XMATCH_PLAN
    version_id = cast(int, plan_config[plan]["version_id"])

    # This is just a sanity check. We need to use version_id=31 and maybe we should
    # just hardcode it.
    assert version_id == 31, "version_id must be 31 to allow using sdss_id."

    too_unmatched = polars.read_database_uri(
        f"""
        SELECT t.* FROM {too_fqtn} t
        LEFT OUTER JOIN {too_rel_fqtn} c2t
        ON (c2t.target_id = t.too_id
            AND c2t.best IS TRUE
            AND c2t.version_id = {version_id})
        WHERE c2t.catalogid IS NULL
        ORDER BY t.too_id
        """,
        database_uri,
        engine="adbc",
    )

    if len(too_unmatched) == 0:
        log.warning("All ToO targets are already matched.")
        return

    # Create the XMatch instance here. We'll need it to get the relational model.
    xmatch_planner = XMatchPlanner.read(
        database,
        plan=plan,
        config_file=TOO_XMATCH_CONFIG,
        log=log,
        log_path=None,
    )

    # Delete temporary tables or fail.
    md5 = xmatch_planner.md5
    for temp_table in ["catalog", "catalog_to_too_target"]:
        temp_table_name = f"{temp_table}_{md5}"
        temp_table_schema = xmatch_planner.temp_schema
        temp_table_full = f"{temp_table_schema}.{temp_table_name}"
        if database.table_exists(temp_table_name, schema=temp_table_schema):
            if overwrite:
                log.warning(f"Dropping table {temp_table_full}.")
                database.execute_sql(f"DROP TABLE {temp_table_full};")
            else:
                raise RuntimeError(f"Table {temp_table_full} already exists.")

    # Step 1: select targets with sdss_id and without catalogid.
    # Populate the catalogid column.
    too_catalogid = too_unmatched.filter(
        polars.col("sdss_id").is_not_null() | polars.col("catalogid").is_not_null()
    ).select("too_id", "sdss_id", "catalogid")

    too_sdss_id_catalogid = polars.read_database_uri(
        f"""
        SELECT t.too_id, s.sdss_id, s.catalogid31 AS catalogid
        FROM {too_fqtn} t
        JOIN catalogdb.sdss_id_stacked s ON (t.sdss_id = s.sdss_id)
        ORDER BY t.too_id
        """,
        database_uri,
        engine="adbc",
    )
    too_sdss_id_catalogid = too_sdss_id_catalogid.filter(
        ~polars.col("catalogid").is_in(too_catalogid["catalogid"])
    )

    too_catalogid = too_catalogid.join(
        too_sdss_id_catalogid[["too_id", "catalogid"]],
        on="too_id",
        how="left",
    )
    too_catalogid = too_catalogid.with_columns(
        catalogid=polars.col.catalogid.fill_null(polars.col.catalogid_right)
    )
    too_catalogid = too_catalogid.select("too_id", "catalogid")

    # Step 2: insert the targets with catalogid into the sanboxed (!)
    # catalog_to_too_target table.
    if len(too_catalogid) > 0:
        rel_model_sb = xmatch_planner.get_relational_model(
            ToO_Target,
            sandboxed=True,
            create=True,
            temp=False,
        )

        rel_model_sb_tn = f"{rel_model_sb._meta.schema}.{rel_model_sb._meta.table_name}"

        log.info(
            f"Adding {len(too_catalogid)} ToO targets with "
            f"catalogid to {rel_model_sb_tn}."
        )

        too_catalogid = too_catalogid.rename({"too_id": "target_id"})
        too_catalogid = too_catalogid.with_columns(
            best=True,
            version_id=polars.lit(version_id, dtype=polars.Int16),
        )

        too_catalogid.write_database(
            rel_model_sb_tn,
            database_uri,
            if_table_exists="append",
            engine="adbc",
        )

        database.execute_sql(f"VACUUM ANALYZE {rel_model_sb_tn};")

    # Step 3: cross-match the remaining targets.
    log.info("Running cross-match for remaining ToO targets.")

    # Set the starting catalogid as the max of the ToO catalogids. This is because
    # we always use run_id=9 for ToOs and we should have plenty of them (and if for
    # some reason we reached the catalogid of run_id=10 it would fail when appending
    # to catalog).
    Catalog = database.models["catalogdb.catalog"]

    # Calculate the range of catalogids for run_id=TOO_RUN_ID.
    min_cid_run = TOO_RUN_ID << (64 - 11)
    max_cid_run = ((TOO_RUN_ID + 1) << (64 - 11)) - 1

    # Get the max catalogid that is between those values.
    max_cid = (
        Catalog.select(peewee.fn.MAX(Catalog.catalogid))
        .where(Catalog.catalogid >= min_cid_run, Catalog.catalogid <= max_cid_run)
        .scalar()
    )

    # If the number exists (it will be None if are no targets for TOO_RUN_ID), set
    # that value as the initial catalogid to assign to new targets.
    if max_cid:
        xmatch_planner._max_cid = max_cid + 1

    xmatch_planner.run(dry_run=dry_run, keep_temp=keep_temp, force=True, vacuum=False)

    return xmatch_planner



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-10
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import pathlib
import shutil
import tempfile
from datetime import datetime

from typing import TYPE_CHECKING, Any

import httpx
import numpy
import polars
import rich.progress

from sdsstools import get_sjd

from too import log
from too.datamodel import too_dtypes
from too.exceptions import ValidationError


if TYPE_CHECKING:
    import rich.console

    from sdssdb.connection import PeeweeDatabaseConnection


__all__ = ["download_file", "read_too_file", "match_fields"]


def read_too_file(
    path: polars.DataFrame | pathlib.Path | str,
    cast: bool = False,
) -> polars.DataFrame:
    """Reads a ToO file in CSV or parquet format.

    Parameters
    ----------
    path
        The path to the file or the dataframe to read.
    cast
        If ``True``, casts the columns to the datamodel types. This may affect
        the precision of the data.

    """

    if isinstance(path, (str, pathlib.Path)):
        path = pathlib.Path(path)

        if path.suffix == ".parquet":
            targets = polars.read_parquet(path)
        elif path.suffix == ".csv":
            targets = polars.read_csv(path)
        else:
            raise ValueError(f"Invalid file type {path.suffix!r}")

    else:
        targets = path

    try:
        if cast:
            targets = targets.cast(too_dtypes)

        targets = targets.sort("added_on", "too_id").select(list(too_dtypes))

    except polars.exceptions.ColumnNotFoundError as err:
        column = str(err).splitlines()[0]
        raise ValidationError(f"Missing column {column!r} in input file.")

    except Exception as err:
        raise ValidationError(f"Error reading input file: {err!r}")

    return targets


def read_all_files(
    path: pathlib.Path | str,
    ignore_invalid: bool = False,
    sort: bool = True,
    silent: bool = False,
) -> polars.DataFrame:
    """Reads all ToO files in a directory.

    Parameters
    ----------
    path
        The path to the directory containing the files. All CSV and Parquet files
        will be read.
    ignore_invalid
        If ``True``, ignores files that cannot be read.
    sort
        If ``True``, sorts the resulting dataframe by ``added_on`` and ``too_id``.
    silent
        If ``True``, does not print any output to the console log.

    """

    path = pathlib.Path(path)

    process_files: list[pathlib.Path] = []
    process_files.extend(path.glob("*.csv"))
    process_files.extend(path.glob("*.parquet"))

    targets = polars.DataFrame({}, schema=too_dtypes)
    for file in process_files:
        try:
            new_targets = read_too_file(file, cast=True)

        except Exception as ee:
            if not silent:
                log.error(f"Failed to read file {file}: {ee}")

            if ignore_invalid:
                continue
            else:
                raise

        else:
            targets = targets.vstack(new_targets)

    if sort:
        targets = targets.sort(["added_on", "too_id"])

    return targets


def deduplicate_too_targets(targets: polars.DataFrame) -> polars.DataFrame:
    """Deduplicates a list of ToO targets preferring the latest ``added_on`` value."""

    targets = targets.sort("added_on", "too_id")
    targets = targets.unique(subset=["too_id"], keep="last")

    return targets


def download_file(
    url: str,
    path: os.PathLike | str,
    transient_progress: bool = False,
    console: rich.console.Console | None = None,
):
    """Downloads a file from a URL to a local path."""

    path = pathlib.Path(path).expanduser()
    path.mkdir(parents=True, exist_ok=True)

    with tempfile.NamedTemporaryFile() as download_file:
        with httpx.stream("GET", url) as response:
            if response.status_code != 200:
                raise RuntimeError(f"URL {url!r} does not exist.")

            total = int(response.headers["Content-Length"])

            with rich.progress.Progress(
                "[progress.percentage]{task.percentage:>3.0f}%",
                rich.progress.BarColumn(bar_width=None),
                rich.progress.DownloadColumn(),
                rich.progress.TransferSpeedColumn(),
                transient=transient_progress,
                console=console,
            ) as progress:
                download_task = progress.add_task("Download", total=total)
                for chunk in response.iter_bytes():
                    download_file.write(chunk)
                    progress.update(
                        download_task,
                        completed=response.num_bytes_downloaded,
                    )

        download_file.flush()
        shutil.move(download_file.name, path / pathlib.Path(url).name)


def match_fields(
    targets: polars.DataFrame,
    database: PeeweeDatabaseConnection,
    rs_version: str | None = None,
    check_separation: bool = False,
) -> polars.DataFrame:
    """Matches a list of targets with their fields and observatories.

    Parameters
    ----------
    targets
        The data frame of targets. It must include fully populated ``ra`` and
        ``dec`` columns.
    database
        The database connection.
    rs_version
        The robostrategy plan to use to select fields. Defaults to the
        value of the ``$RS_VERSION`` environment variable.
    check_separation
        If ``True``, checks that the separation between the target and the
        field centre is less than the FPS FoV.

    Returns
    -------
    dataframe
        The input dataframe with ``field_id`` and ``observatory`` columns.

    """

    from astropy.coordinates import SkyCoord, match_coordinates_sky

    from coordio.defaults import APO_MAX_FIELD_R, LCO_MAX_FIELD_R

    targets = targets.clone()

    if rs_version is None:
        rs_version = os.environ.get("RS_VERSION", None)
        if rs_version is None:
            raise ValueError("No rs_version provided and $RS_VERSION not set.")

    too_sc = SkyCoord(
        ra=targets["ra"].to_numpy(),
        dec=targets["dec"].to_numpy(),
        unit="deg",
        frame="icrs",
    )

    fields = polars.read_database(
        "SELECT f.field_id, f.racen AS field_ra, "
        "       f.deccen AS field_dec, o.label AS observatory "
        "FROM targetdb.field f "
        "JOIN targetdb.version v ON f.version_pk = v.pk "
        "JOIN targetdb.observatory o ON o.pk = f.observatory_pk "
        f"WHERE v.plan = '{rs_version}' AND v.robostrategy;",
        database,
    )

    if len(fields) == 0:
        raise ValueError(f"No fields found for rs_version={rs_version!r}")

    fields_sc = SkyCoord(
        ra=fields["field_ra"].to_numpy(),
        dec=fields["field_dec"].to_numpy(),
        unit="deg",
        frame="icrs",
    )

    field_idx, sep2d, _ = match_coordinates_sky(too_sc, fields_sc)

    field_to_target = fields[field_idx]
    field_to_target = field_to_target.with_columns(field_separation=sep2d.deg)

    targets = targets.hstack(field_to_target)

    if check_separation:
        for obs, max_field_r in [("APO", APO_MAX_FIELD_R), ("LCO", LCO_MAX_FIELD_R)]:
            obs_targets = targets.filter(polars.col.observatory == obs)
            if (obs_targets["field_separation"] > max_field_r).any():
                raise ValueError(
                    f"Targets with separation larger than {max_field_r} deg "
                    f"found for observatory {obs}."
                )

    return targets


def create_mock_catalogue(
    data: polars.DataFrame | dict[str, Any],
) -> polars.DataFrame:  # pragma: no cover
    """Tool to generate fake ToO input files with correct values."""

    if isinstance(data, dict):
        data = polars.DataFrame(data)

    assert isinstance(data, polars.DataFrame)

    for column in too_dtypes:
        if column not in data.columns:
            if column == "too_id":
                col_data = numpy.arange(len(data))
            elif column == "fiber_type":
                col_data = ["APOGEE"] * len(data)
            elif column == "sky_brightness_mode":
                col_data = ["bright"] * len(data)
            elif column == "ra":
                col_data = numpy.random.uniform(0, 360, len(data))
            elif column == "dec":
                col_data = numpy.random.uniform(-90, 90, len(data))
            elif column == "can_offset":
                col_data = [False] * len(data)
            elif column == "n_exposures":
                col_data = [1] * len(data)
            elif column == "priority":
                col_data = [5] * len(data)
            elif column == "active":
                col_data = [True] * len(data)
            elif column == "observe_from_mjd":
                col_data = [get_sjd("APO") - 1] * len(data)
            elif column == "observe_until_mjd":
                col_data = [get_sjd("APO") + 10] * len(data)
            elif column == "observed":
                col_data = [False] * len(data)
            elif column == "added_on":
                col_data = [datetime.now().isoformat().split("T")[0]] * len(data)
            else:
                col_data = [None] * len(data)

            data = data.with_columns(
                **{column: polars.Series(col_data, dtype=too_dtypes[column])}
            )

    data = data.cast(too_dtypes).select(too_dtypes.keys()).sort("added_on", "too_id")

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-12
# @Filename: exceptions.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations


__all__ = ["TooError", "ValidationError"]


class TooError(Exception):
    """A base class for exceptions in the ``too`` package."""

    pass


class ValidationError(TooError):
    """An exception raised when ToO target validation fails."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-07
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import warnings

import sdssdb
from sdsstools import get_logger, get_package_version


sdssdb.autoconnect = False  # type: ignore


NAME = "sdss-too"


log = get_logger(NAME, use_rich_handler=True)

console = log.rich_console
assert console is not None


__version__ = get_package_version(path=__file__, package_name=NAME)


from .carton import *
from .database import *
from .datamodel import *
from .dump import *
from .mock import *
from .tools import *
from .validate import *
from .xmatch import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-02-08
# @Filename: mock.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib

from typing import Literal, overload

import numpy
import polars

from too import console, log
from too.datamodel import too_dtypes
from too.tools import download_file


__all__ = ["create_mock_too_catalogue"]


@overload
def get_sample_data(
    table: Literal["gaia_dr3", "twomass", "photoobj"],
    lazy: Literal[True],
) -> polars.LazyFrame: ...


@overload
def get_sample_data(
    table: Literal["gaia_dr3", "twomass", "photoobj"],
    lazy: Literal[False],
) -> polars.DataFrame: ...


def get_sample_data(
    table: Literal["gaia_dr3", "twomass", "photoobj"],
    lazy: bool = True,
) -> polars.DataFrame | polars.LazyFrame:
    """Downloads and caches the table sample data.

    Parameters
    ----------
    table
        The table to download. Can be ``'gaia_dr3'`` or ``'twomass'``. If the table
        file is not found in the cache, it will be downloaded from the SAS server.
    lazy
        Whether to return a lazy dataframe.

    Returns
    -------
    A polars dataframe with the sample data.

    """

    BASE_URL = "https://data.sdss5.org/resources/target/mocks/samples/"
    CACHE_PATH = pathlib.Path("~/.cache/sdss/too/samples").expanduser()

    if table == "gaia_dr3":
        filename = "gaia_dr3_1M_sample.parquet"
    elif table == "twomass":
        filename = "twomass_psc_1M_sample.parquet"
    elif table == "photoobj":
        filename = "sdss_dr13_photoobj_1M_sample.parquet"
    else:
        raise ValueError(f"Invalid table {table!r}")

    if not (CACHE_PATH / filename).exists():
        log.debug(f"File {filename!r} not found in cache. Downloading from {BASE_URL}.")
        download_file(
            f"{BASE_URL}/{filename}",
            CACHE_PATH,
            transient_progress=True,
            console=console,
        )

    if lazy:
        return polars.scan_parquet(CACHE_PATH / filename)
    return polars.read_parquet(CACHE_PATH / filename)


def get_sample_targets(
    table: Literal["gaia_dr3", "twomass", "photoobj"],
    n_targets: int,
    seed: int | None = None,
):
    """Returns a sample of table targets."""

    data = get_sample_data(table, lazy=False)
    sample = data.sample(n_targets, seed=seed, shuffle=True)

    if table == "gaia_dr3":
        col_mapping = {
            "source_id": "gaia_dr3_source_id",
            "ra": "ra",
            "dec": "dec",
            "pmra": "pmra",
            "pmdec": "pmdec",
            "parallax": "parallax",
            "phot_g_mean_mag": "gaia_g_mag",
            "phot_bp_mean_mag": "gaia_bp_mag",
            "phot_rp_mean_mag": "gaia_rp_mag",
        }
    elif table == "twomass":
        col_mapping = {
            "pts_key": "twomass_pts_key",
            "ra": "ra",
            "decl": "dec",
            "h_m": "h_mag",
        }
    elif table == "photoobj":
        col_mapping = {
            "ra": "ra",
            "dec": "dec",
            "psfmag_u": "u_mag",
            "psfmag_g": "g_mag",
            "psfmag_r": "r_mag",
            "psfmag_i": "i_mag",
            "psfmag_z": "z_mag",
        }

    if "source_id" in sample.columns:
        sample = sample.cast({"source_id": polars.Int64})
    elif "pts_key" in sample.columns:
        sample = sample.cast({"pts_key": polars.Int32})

    sample = sample.select(*list(col_mapping), "catalogid", "sdss_id")
    sample = sample.rename(col_mapping)
    sample = sample.cast({"catalogid": polars.Int64, "sdss_id": polars.Int64})

    sample = sample.with_columns(polars.col(["ra", "dec"]).cast(polars.Float64))

    return sample


def create_mock_too_catalogue(
    n_targets: int = 1000000,
    fraction_unknown: float = 0.6,
    fraction_unknown_sdss: float = 0.2,
    fraction_unknown_gaia: float = 0.3,
    fraction_known_gaia: float = 0.8,
    fraction_known_sdss: float = 0.1,
    fraction_known_twomass: float = 0.1,
    catalogid_likelihood: float = 0.2,
    design_modes: list[str] = ["bright", "dark"],
    seed: int | None = None,
):
    """Creates a mock ToO catalogue.

    Parameters
    ----------
    n_targets
        The number of targets in the mocked catalogue.
    fraction_unknown
        Fraction of the targets that will not have an associated target in
        ``catalogdb``.
    fraction_unknown_sdss : float
        Fraction of the unknown targets that will actually be drawn from
        ``sdss_dr13_photoobj``.
    fraction_unknown_gaia : float
        Fraction of the unknown targets that will actually be drawn from
        ``gaia_dr3_source``.
    fraction_known_gaia
        Fraction of the targets that will have a known Gaia source. Along with
        ``fraction_known_sdss`` and ``fraction_known_twomass``, must add up to 1.
        The total number of targets with ``gaia_dr3_source_id`` will be
        ``n_targets * fraction_known_gaia * (1 - fraction_unknown)``.
    fraction known_sdss
        Fraction of the targets that will have a known SDSS source.
    fraction_known_twomass
        Fraction of the targets that will have a known 2MASS source.
    catalogid_likelihood
        The likelihood (0-1) that a target will have a known ``catalogid`` or
        ``sdss_id``.
    design_modes
        A list of design modes to randomly assign to the ``sky_brightness_mode``
        column.
    seed
        An optional random seed to use.

    Returns
    -------
    A polars dataframe with the mock catalogue.

    """

    n_known = n_targets * (1 - fraction_unknown)

    gaia_known_targets = get_sample_targets(
        "gaia_dr3",
        int(n_known * fraction_known_gaia),
        seed=seed,
    )

    twomass_known_targets = get_sample_targets(
        "twomass",
        int(n_known * fraction_known_twomass),
        seed=seed,
    )

    sdss_known_targets = get_sample_targets(
        "photoobj",
        int(n_known * fraction_known_sdss),
        seed=seed,
    )
    sdss_known_targets = sdss_known_targets.with_columns(
        catalogid=None,
        sdss_id=None,
        seed=seed,
    )

    n_unknown = n_targets * fraction_unknown

    gaia_unknown_targets = get_sample_targets(
        "gaia_dr3",
        int(n_unknown * fraction_unknown_gaia),
        seed=seed,
    )
    gaia_unknown_targets = gaia_unknown_targets.with_columns(
        gaia_dr3_source_id=None,
        catalogid=None,
        sdss_id=None,
        gaia_bp_mag=None,
        gaia_rp_mag=None,
    )

    sdss_unknown_targets = get_sample_targets(
        "photoobj",
        int(n_unknown * fraction_unknown_sdss),
        seed=seed,
    )
    sdss_unknown_targets = sdss_unknown_targets.with_columns(
        catalogid=None,
        sdss_id=None,
    )

    df = polars.DataFrame(schema=too_dtypes)
    df = polars.concat(
        [
            df,
            gaia_known_targets,
            twomass_known_targets,
            sdss_known_targets,
            gaia_unknown_targets,
            sdss_unknown_targets,
        ],
        how="diagonal",
    )

    n_random = int(n_targets - df.height)

    # Give half of them Gaia mags and half SDSS.
    random_gaia = [None] * n_random
    random_sdss = [None] * n_random
    random_gaia[: n_random // 2] = numpy.random.uniform(10, 20, size=n_random // 2)
    random_sdss[n_random // 2 :] = numpy.random.uniform(10, 20, size=n_random // 2)

    random_targets = polars.DataFrame(
        {
            "ra": numpy.random.uniform(0, 360, n_random),
            "dec": numpy.random.uniform(-90, 90, n_random),
            "gaia_g_mag": polars.Series(random_gaia, dtype=polars.Float32),
            "u_mag": polars.Series(random_sdss, dtype=polars.Float32),
            "g_mag": polars.Series(random_sdss, dtype=polars.Float32),
            "r_mag": polars.Series(random_sdss, dtype=polars.Float32),
            "i_mag": polars.Series(random_sdss, dtype=polars.Float32),
            "z_mag": polars.Series(random_sdss, dtype=polars.Float32),
        },
    )

    df = polars.concat([df, random_targets], how="diagonal")

    df = df.with_columns(
        keep_cid=polars.Series(numpy.random.rand(df.height) < catalogid_likelihood)
    )
    df = df.with_columns(
        catalogid=polars.when(polars.col.keep_cid).then(polars.col.catalogid),
        sdss_id=polars.when(polars.col.keep_cid).then(polars.col.sdss_id),
    )
    df.drop_in_place("keep_cid")
    df = df.sample(df.height, shuffle=True, seed=seed)  # Shuffle

    df = df.with_columns(
        polars.int_range(1, df.height + 1).alias("too_id"),
        program=polars.lit("TNS", dtype=polars.String),
    )

    # Drop column "seed"
    if "seed" in df.columns:
        df = df.drop("seed")

    # Fill out some columns
    fiber_type = numpy.array(["APOGEE"] * df.height)
    boss_mask = numpy.random.rand(df.height) < 0.5
    fiber_type[numpy.where(boss_mask)[0]] = "BOSS"
    sky_brightness_mode = numpy.random.choice(design_modes, size=df.height)

    df = df.with_columns(
        fiber_type=polars.Series(fiber_type, dtype=polars.String),
        observed=False,
        can_offset=True,
        active=True,
        priority=polars.lit(5, dtype=polars.Int16),
        n_exposures=polars.lit(3, dtype=polars.Int16),
        sky_brightness_mode=polars.Series(sky_brightness_mode, dtype=polars.String),
    )

    return df
